# -*- coding: utf-8 -*-
require 'Backend/datatypes'

class SymLexer
  attr_reader :sym_code

  def initialize(sym_code, debug_mode = false)
    @debug_mode = debug_mode
    @sym_code = sym_code
    @line_code = ""              # Stores the code from the current line

    # All custom datatypes (classes)
    @datatypes = []

    # If set to true, will return last lexem on the next get_nex_lexem() call
    @return_previous = false
    @previous_lexem = nil        # 1 Step back
    @grand_previous_lexem = nil  # 2 Steps back
    generate_regexp()
  end

  # Gets the remaining code with excessive whitespace removed
  # This is used for debug output
  def code_left(code = @sym_code)
    re = /\s+/
    return code.gsub(re, " ").strip
  end

  # Returns an appropiate chunk of code to use in errors
  def get_line()
    return @line_code + /.*/.match(@sym_code)[0]
  end
  # Empties @line_code variable, (called when change line)
  def flush_line()
    @line_code = ""
  end

  def add_datatype(new_type)
    puts "Lexer: Adding new datatype #{new_type}" if @debug_mode
    @datatypes << new_type  
    generate_regexp()
  end

  # Returns function body (also used for methods as they
  #                         are only functions in a confined scope)
  def get_function_body()
    total_re = //m
    any_re = '.*?'
    end_re = '\\\\' + '\$'
    begins = 0
    endloop = true

    result = @sym_code
    begin
      endloop = true
      # (any + (end+any)*begins) + end
      total_re = Regexp.new(any_re + (end_re + any_re)*begins + end_re ,
                                                            Regexp::MULTILINE)
      result = total_re.match(@sym_code)
      # If there was no match, the syntax was wrong
      raise "SYM-ERROR: Invalid functiondef at #{get_line()},"+
                  " proper syntax is: \n\$\/function()\n\\\$" if not result
      
      amount = result[0].split("\$\/").length-1
      if amount != begins
        endloop = false
        begins = amount
      end
    end while (not endloop)

    @sym_code = result.post_match()
    puts "Lexer: Matched function body. Left= #{code_left}" if @debug_mode
    return result[0].slice(0..-3) # Slice off the end-tag
  end
  
  def get_code_body(parts)
    local_debug = false
    puts "@sym_code before:\n\"#{@sym_code.to_s}\"" if local_debug

    condition = code = next_part = nil

    re_condition = /(.*)/
    condition_match = re_condition.match(@sym_code)

    if (condition_match)
      condition = condition_match.captures[0]
      @sym_code = condition_match.post_match
      re_code = /(.*?)(#{parts})/m
      match_code = re_code.match(@sym_code)
      if (match_code)
        code = match_code.captures[0]
        next_part = match_code.captures[1]
        @sym_code = match_code.post_match
      end
    end

    puts "@sym_code after:\n\"#{@sym_code.to_s}\"" if local_debug
    condition = condition.strip
    return [code, condition, next_part]
  end
  
  def add_regexp(regexp, datatype)
    @regexp << [regexp, Proc.new {|x| Lexem.new(datatype, x)}]
  end

  # Generates a list of regexp's. In pairs as follows:
  # 1. Regexp, the matching pattern
  # 2. A Code block(Proc) that creates the lexem object when called
  def generate_regexp()
    def add_regexp(regexp, datatype)
      @regexp << [regexp, Proc.new {|x| Lexem.new(datatype, x)}]
    end
    
    puts "Lexer: Re-generating all regexp" if @debug_mode
    @regexp = []

    # Multiline comment
    @regexp << [/\A\#\/.*?\\\#/m, false]
    # Comment
    @regexp << [/\A\#.*$/, false]
    # Space or Tab
    @regexp << [/\A(\ |\t)+/, false]	
	
    # If Parts
    add_regexp(/\A\?\//, IfPart)
    add_regexp(/\A\\\?\//, ElseIfPart)
    add_regexp(/\A\\\*/, ElsePart)
    add_regexp(/\A\\\?/, EndIfPart)

    # For Parts
    add_regexp(/\A\.\.\//, ForPart)
    add_regexp(/\A\\\.\./, EndForPart)

    # While Parts
    add_regexp(/\A\|\//, WhilePart)
    add_regexp(/\A\\\|/, EndWhilePart)

    # For Parts
    add_regexp(/\A\.\.\/\s+\d+\s+\d+\s+[a-z][a-zA-Z0-9_]*/, ForPart)
    add_regexp(/\A\\\.\./, EndForPart)

    # While Parts
    add_regexp(/\A\|\//, WhilePart)
    add_regexp(/\A\\\|/, EndWhilePart)

    # Break
    add_regexp(/\A\<\-\-/, BreakOperator)

    # IncrementOperator
    add_regexp(/\A[a-z][a-zA-Z0-9\_]*\+\+/, IncrementOperator)


    # Builtin datatypes
    @regexp << [/\AInt/, 
            Proc.new {|x| Lexem.new( DataType, DataType.new(Int))}]
    @regexp << [/\AFixnum/, 
            Proc.new {|x| Lexem.new( DataType, DataType.new(Fixnum))}]
    @regexp << [/\AFloat/,
            Proc.new {|x| Lexem.new( DataType, DataType.new(Float))}]
    @regexp << [/\AString/,
            Proc.new {|x| Lexem.new( DataType, DataType.new(String))}]
    @regexp << [/\AVoid/,
            Proc.new {|x| Lexem.new( DataType, DataType.new(Void))}]
    
    # Classes
    for d_type in @datatypes
      @regexp << [/\A#{d_type}\.init\(.*\)/,
             Proc.new {|x| Lexem.new(ClassInit, x)}]
      @regexp << [/\A#{d_type.to_s}/,
             Proc.new {|x| Lexem.new( DataType, DataType.new(d_type))}]
    end

    # Comma (in function-def, lists etc..)
    @regexp << [/\A\,/,
            Proc.new{|x| Lexem.new(Comma, Comma.new())}]
    
    # Function definition
    @regexp << [/\A\$\//,
            Proc.new {|x| Lexem.new( FunctionDef, FunctionDef.new(false))}]

    # Class definition
    @regexp << [/\A\@\/[A-Z][a-z0-9\_]*/,
            Proc.new {|x| Lexem.new( ClassDef, x)}]
    @regexp << [/\A\\\@/,
            Proc.new {|x| Lexem.new( ClassEnd, "")}]
    # Return statement
    @regexp << [/\Areturn/,
            Proc.new {|x| Lexem.new( ReturnOperator, "")}]

    # Print-operator without newline
    @regexp << [/\A\<\<\</,
            Proc.new {|x| Lexem.new(PrintOperator, true)}]
    # Print-operator with newline
    @regexp << [/\A\<\</,
            Proc.new {|x| Lexem.new(PrintOperator, false)}]

    # Input-operator
    @regexp << [/\A\>\>/,
            Proc.new {|x| Lexem.new(InputOperator, x)}]

    # Modifier-operators
    @regexp << [/\A(\+\=|\-\=|\*\=|\/\=|\%\=|\*\*\=|\^\=)/,
            Proc.new {|x| Lexem.new(ModifierOperator, x)}]
    @regexp << [/\A\-\=/,
            Proc.new {|x| Lexem.new(ModifierOperator, x)}]
    # Comparison-operators
    @regexp << [/\A\>\=/,
            Proc.new {|x| Lexem.new(Operator, x)}]
    @regexp << [/\A\<\=/,
            Proc.new {|x| Lexem.new(Operator, x)}]
    @regexp << [/\A[\<\>]/,
            Proc.new {|x| Lexem.new(Operator, x)}]

    # Logical operators
    add_regexp(/\A\&\&/, Operator)
    add_regexp(/\A\|\|/, Operator)
    # Equality operator
    add_regexp(/\A\=\=/, Operator)
    # Negate equality operator
    add_regexp(/\A\!\=/, Operator)
    # Equality operator
    add_regexp(/\A\=\=/, Operator)
    # Negate equality operator
    add_regexp(/\A\!\=/, Operator)

    # Assignment
    @regexp << [/\A\=/,
            Proc.new{|x| Lexem.new(AssignmentOperator, x)}]
    # Float
    @regexp << [/\A\d+\.\d+/,
            Proc.new {|x| Lexem.new(Float, x.to_f)} ]
    # Negative Float
    @regexp << [/\A-\d+\.\d+/,
            Proc.new {|x| Lexem.new(Float, x.to_f)} ]
    # Integer
    @regexp << [/\A\d+/,
            Proc.new {|x| Lexem.new(Integer, x.to_i)} ]
    # Negative Integer
    @regexp << [/\A-\d+/,
            Proc.new {|x| Lexem.new(Integer, x.to_i)} ]

    # List/Dict indexing
    @regexp << [/\A[a-z][a-zA-Z0-9\_]*\[.+\]/,
            Proc.new {|x| Lexem.new(Indexing, x)} ]

    # List start
    @regexp << [/\A\[/,
            Proc.new {|x| Lexem.new(ListStart, x)} ]
    # List end
    @regexp << [/\A\]/,
            Proc.new {|x| Lexem.new(ListEnd, x)} ]

    # Dict relation, used to seperate each element
    @regexp << [/\A\:/,
            Proc.new {|x| Lexem.new(DictRelation, x)} ]

    # Dict start
    @regexp << [/\A\{/,
            Proc.new {|x| Lexem.new(DictStart, x)} ]

    # Dict end
    @regexp << [/\A\}/,
            Proc.new {|x| Lexem.new(DictEnd, x)} ]

    # Operators
    @regexp << [/\A[\*]{2}/x,
            Proc.new {|x| Lexem.new(Operator, x)}]
    @regexp << [/\A[\+\*\/\-\^\%]/x,
            Proc.new {|x| Lexem.new(Operator, x)}]

    # Terminator (newline)
    @regexp << [/\A(\r|\n)+/,
            Proc.new{|x| Lexem.new(Terminator, "Newline")}]

    # String
    @regexp << [/\A\".*?\"/m,
            Proc.new{|x| Lexem.new(String, x)}]

    # Method call
    @regexp << [/\A[a-z][a-zA-Z0-9\_]*\.\w+\((( )*\-?\"?[0-9A-Za-z]+\.?[0-9]*\"?)?(,( )*\-?\"?[0-9A-Za-z]+\.?[0-9]*\"?)*( )*\)/,
            Proc.new {|x| Lexem.new( MethodCall, x)}]
    # Function call
    @regexp << [/\A\w+\((( )*\-?\"?[0-9A-Za-z]+\.?[0-9]*\"?)?(,( )*\-?\"?[0-9A-Za-z]+\.?[0-9]*\"?)*( )*\)/,
            Proc.new {|x| Lexem.new(FunctionCall, x)}]

    # Variable name
    @regexp << [/\A[a-z][a-zA-Z0-9\_]*/,
            Proc.new{|x| Lexem.new(Variable, x)}]
    # Constant name
    @regexp << [/\A[A-Z][A-Z0-9\_]*/,
            Proc.new{|x| Lexem.new(Constant, x)}]

    # LParanthesis
    @regexp << [/\A\(/,
            Proc.new {|x| Lexem.new(LParanthesis, "(")}]
    # RParanthis
    @regexp << [/\A\)/,
            Proc.new {|x| Lexem.new(RParanthesis, ")")}]

    # Syntax error catching
    # Function end tag
    @regexp << [/\A\\\$/, 
            Proc.new {|x| raise "SYM-ERROR #{get_line()}:" +
                                               " Unmatched function end tag"}]
  end

  # Used by SymParser, returns a Lexem-object
  # Iterates over all rules generated by generate_regexp()
  # If there's a match on the regexp, the Proc is run
  def get_next_lexem()
    puts "Lexer: Getting a lexem code left='#{code_left()}'" if @debug_mode
    if @return_previous
      puts "Lexer: Returning previous ( #{@previous_lexem.val}, "+
                        "it's a #{@previous_lexem.type} )" if @debug_mode
      @return_previous = false
      return @previous_lexem
    end

    if @sym_code.length == 0
      puts "Lexer: Out of code, returning EOF" if @debug_mode
      return Lexem.new(EOF, 0)
    end

    error = true
    # Search through all regexp's and look for a match
    for i in 0..@regexp.length-1
      match = @regexp[i][0].match(@sym_code)

      if (match)
        @sym_code = match.post_match
        @line_code += match[0]
        puts "Lexer: Match on regexp for #{@regexp[i][1].call.type if
             @regexp[i][1]}=#{@regexp[i][0]} code left=#{code_left()}" if
                                                          @debug_mode and
                                                                @regexp[i][1]

        # If there is a block to call, call it
        if (@regexp[i][1])
          lexem = @regexp[i][1].call(match.to_s)
          error = false
        else 
          # Otherwise re-call ourselves at the new position
          return self.get_next_lexem()
        end
        break
      end #end match
    end # end for

    # If there was an error, print everything from this point until next line
    raise "SYM-ERROR: Invalid syntax, " +
                              "could not recognize: #{get_line()}\n" if error
    @grand_previous_lexem = @previous_lexem
    @previous_lexem = lexem

    puts "Lexer: Returning #{lexem.val}, it's a #{lexem.type}" if
                                      @debug_mode and lexem.type != Terminator
    puts "Lexer: Code left: #{code_left()}" if @debug_mode
    puts "----" if @debug_mode
    return lexem
  end

  def return_previous()
    puts "Lexer: Return previous-flag has been set" if @debug_mode
    @return_previous = true
  end

  def return_next()
    puts "Lexer: Disabling previous flag" if @debug_mode
    @return_previous = false
  end

  def get_grand()
    return @grand_previous_lexem
  end

end
