# -*- coding: utf-8 -*-
require 'Backend/lexer'
require 'Backend/scope'
class SymParser
  def initialize(debug_mode = false, silent_mode = false)
    @debug_mode = debug_mode
    @silent_mode = silent_mode    # Print does nothing if silent_mode is true
    @lexer = nil
    
    # All scopes, last one is "more local"
    @scopes = []
    # Create global scope
    @scopes << Scope.new({"NULL" => nil, 
                          "SYM_VERSION" => "1.0",
                          "sym_max_recursion"=>100,       # Max recursion depth
                          "sym_current_recursion"=>0,     # Current depth
                          "true"=>true,
                          "false"=>false
                          }, @debug_mode)

    @in_fun = nil         # Reference to current function we're currently in
    @return_level = 0     # Keeps track of scope depth
    @return_value = 0     # When a function returns, the value is saved here
    @break = false        # True=Break out of parsing ASAP

    @break_loop = false
    @loop_level = 0				# Keeps track of loop depth
    @profile_mode = false	# Profile code and create report
  end

  # Input is Sym code, as string
  # Also used to call functions, methods etc.
  def parse(input)
    # Save the old lexer (we need to be able to return after call)
    old_lexer = @lexer

    # Start iterating over the new code until end of file
    @lexer = SymLexer.new(input, @debug_mode)
    while evaluate_statement()
    end

    @lexer = old_lexer
  end

private
  # Searches all scopes for an object and returns its _name_
  # Argument is the object to be found, as actual reference
  def find_name(object)
    puts "Parser: Trying to find #{object}" if @debug_mode

    # Iterate from back and up
    for i in (@scopes.length-1).downto(0)
        obj = @scopes[i].get_object_name(object)
        return obj if obj != nil # If found, return it
    end
    raise "SYM-ERROR: #{name} not found"
  end

  # Searches all scopes for an object and returns its _reference_
  # type is only used as debug message
  # if raise_error=true, an error will be raised if the object isn't found
  def seek_storage(name, type = Variable, raise_error = true)
    puts "Parser: seek_storage. Trying to find #{type} #{name}" if @debug_mode

    # If we are in a function, use it's lexical scope instead of dynamic scope
    if @in_fun != nil
      old_scopes = @scopes
      @scopes = @in_fun.get_scope_ref()
      @scopes << old_scopes[-1]         # Also add the function local scope

#      for sc in @scopes
#        puts sc.to_s
#      end
#      puts "-----------------"
    end

    # Iterate from back and up
    for i in (@scopes.length-1).downto(0)
        obj = @scopes[i].get_name(name)

        @return_level = i
        # When we return the level-pointer will be invalid because the old
        # scope is restored, add the length-difference to compensate
        @return_level += (old_scopes.length - @scopes.length) if @in_fun != nil

        # Return to the previous scope if we are in a function
        @scopes = old_scopes if obj != nil and @in_fun != nil
        return obj if obj != nil # If found, return it
    end
    
    raise "SYM-ERROR #{@lexer.get_line()}: #{name} not found" if raise_error
    return false
  end

  # Evaluates statement and runs the apppropiate action
  def evaluate_statement()
    puts "Parser: Entering evaluate_statment()" if @debug_mode

    # Break from function/method
    if @break
      @break = false
      return false
    end

    # Break from loop
    if @break_loop
      return false
    end

    @lexer.return_next()
    lexem = @lexer.get_next_lexem()

    if lexem.type == EOF
      return false
    elsif lexem.type == PrintOperator
      print_operator(lexem)
    elsif lexem.type == InputOperator
      input_operator()
    elsif lexem.type == ReturnOperator
      return_operator()
    elsif lexem.type == AssignmentOperator
      assignment_operator()
    elsif lexem.type == ClassDef
      class_def(lexem)
    elsif lexem.type == FunctionDef
      # If it was an end-def, raise error
      raise "SYM-ERROR #{@lexer.get_line()}: Unmatched function-end" if
                                                                  lexem.val.end
      ans = function_def()
      name = ans[0]
      ret = ans[1]
      code = ans[2]
      args = ans[3]
      scope_ref = ans[4]
      obj = Function.new(name, ret, code, args, scope_ref)
      @scopes[-1].store_object(name, obj)
    elsif lexem.type == Terminator
      # Flush "this line" buffer in lexer
      @lexer.flush_line()
      # evaluate next statement
      evaluate_statement()
    elsif lexem.type == Variable or lexem.type == Constant
      l = @lexer.get_next_lexem()
      if l.type == AssignmentOperator
        assignment_operator()
      elsif l.type == ModifierOperator
        modifier_operator(l.val)
      elsif l.type == MethodCall
        call_method(l, lexem.val)
      else
        if lexem.type == Constant and l.type == Variable
          raise "SYM-ERROR #{@lexer.get_line()}: Invalid identifier. First letter of a variable must be lower case. Constants must be all upper case"
        else
          raise "SYM-ERROR #{@lexer.get_line()}: Invalid statement " +
                      "#{lexem.val}"
        end

      end
    elsif lexem.type == FunctionCall
      call_function(lexem)
    elsif lexem.type == ClassInit
      class_init(lexem)
    elsif lexem.type == MethodCall
      r = /(.+)\./
      match = r.match(lexem.val)
      raise "SYM-ERROR #{@lexer.get_line()}: Invalid " +
                                      "methodcall. #{lexem.val}" if not match
      call_method(lexem, match[1])

    elsif lexem.type == ModifierOperator
      modifier_operator(lexem.val)
    elsif lexem.type == IfPart
      conditional_statement()
    elsif lexem.type == ForPart
      iterative_for_statement()
    elsif lexem.type == WhilePart
      iterative_while_statement()
    elsif lexem.type == BreakOperator
      break_operator()
    elsif lexem.type == IncrementOperator
      increment_operator(lexem.val)
    elsif lexem.type == DictEnd
      raise "SYM-ERROR #{@lexer.get_line()}: Unmatched dictionary end, }"
    elsif lexem.type == ListEnd
      raise "SYM-ERROR #{@lexer.get_line()}: Unmatched list end, ]"
    elsif lexem.type == Comma
      raise "SYM-ERROR #{@lexer.get_line()}: Unknown comma, " +
                                            "likely from an unmatched list"
    else
      raise "SYM-ERROR #{@lexer.get_line()}: " +
                                       "Invalid statement."
    end

    return true
  end

  def increment_operator(lexem_val)
    local_debug = false

    puts "Parser: Entering increment_operator()" if @debug_mode or local_debug
    lexem_val = lexem_val.slice(0, lexem_val.length-2)
    found_it = @scopes[-1].get_name(lexem_val)
    begin
      Float(found_it)
    rescue # If value can't be floated, it can't be a number
      raise "SYM-ERROR: #{@lexer.get_line().strip()}: Not a number"
    else
      # Is numeric
      found_it += 1
      @scopes[-1].store_object(lexem_val, found_it)
    end
  end

 def calculate_expr()
  puts "Parser: Entering calculate_expr()" if @debug_mode
  factor1 = calculate_factor()

  lexem = @lexer.get_next_lexem()
  while lexem.val == "+" or lexem.val == "-"
    factor2 = calculate_factor()

    if lexem.val == "+"
      factor1 += factor2
    else
      factor1 -= factor2
    end

    lexem = @lexer.get_next_lexem()
  end
  @lexer.return_previous()

  return factor1
end

 def calculate_factor()
    puts "Parser: Entering calculate_factor()" if @debug_mode
    factor1 = get_number()

    lexem = @lexer.get_next_lexem()

    allowed_operators = ["*", "/", "**", "^", "%", "<", ">", "<=", ">=", "==", "!=", "||", "&&"]
	while allowed_operators.include?(lexem.val)
      factor2 = get_number()
      #return factor2 if factor2.type == SymObject

      if lexem.val == "*"
		raise "Operation not allowed" if factor2.class == String
        factor1 *= factor2
      elsif lexem.val == "**" or lexem.val == "^"
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 **= factor2
      elsif lexem.val == "/"
        raise "SYM-ERROR #{@lexer.get_line()}: Division by zero" if factor2.ceil == 0
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 /= factor2
      elsif lexem.val == "%"
		check_numeric_error(factor1, factor2, lexem.val)
		raise "SYM-ERROR #{@lexer.get_line()}: Division by zero" if factor2.ceil == 0
        factor1 %= factor2
      elsif lexem.val == "<"
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 = factor1 < factor2
      elsif lexem.val == "<="
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 = factor1 <= factor2
      elsif lexem.val == ">="
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 = factor1 >= factor2
      elsif lexem.val == ">"
		check_numeric_error(factor1, factor2, lexem.val)
        factor1 = factor1 > factor2
	  elsif lexem.val == "=="
      factor1 = (factor1 == factor2)
	  elsif lexem.val == "!="
      factor1 = (factor1 != factor2)
	   elsif lexem.val == "||"
      factor1 = 1 if factor1 == true
      factor1 = 0 if factor1 == false
      factor2 = 1 if factor2 == true
      factor2 = 0 if factor2 == false
      factor1 = not(factor1 == 0 and factor2 == 0)
    elsif lexem.val == "&&"
      factor1 = factor1 and factor2
	  elsif lexem.val == "=="
      factor1 = (factor1 == factor2)
	  elsif lexem.val == "!="
      factor1 = (factor1 != factor2)
    end

      lexem = @lexer.get_next_lexem()
    end
    @lexer.return_previous()

    return factor1
  end

  def get_number()
    puts "Parser: Entering get_number()" if @debug_mode
    lexem = @lexer.get_next_lexem()

    if lexem.type == FunctionCall
      return call_function(lexem)
    elsif lexem.type == ListStart
      return get_list(lexem)
    elsif lexem.type == ListEnd
      return lexem
    elsif lexem.type == Indexing
      return get_by_index(lexem)
    elsif lexem.type == DictStart
      return get_dict(lexem)
    elsif lexem.type == DictEnd
      return lexem
    elsif lexem.type == ClassInit
      return class_init(lexem)

    elsif lexem.type == MethodCall
      r = /(.+)\./
      match = r.match(lexem.val)
      raise "SYM-ERROR #{@lexer.get_line()}: Invalid " +
                                    "methodcall. #{lexem.val}" if not match
      return call_method(lexem, match[1])

    elsif lexem.type == LParanthesis
      value = calculate_expr()

      # Expected right-paranthesis
      er = @lexer.get_next_lexem()
      raise "SYM-ERROR #{@lexer.get_line()}: Unbalanced paranthesis got #{er.type}" if er.type != RParanthesis
    elsif lexem.type == Integer or lexem.type == Float
      value = lexem.val
    elsif lexem.type == String
      value = lexem.val
    elsif lexem.type == Variable
      value = seek_storage(lexem.val)
    elsif lexem.type == Constant
      value = seek_storage(lexem.val, Constant)
    elsif lexem.type == Terminator
      return lexem
    elsif lexem.type == MethodCall
      method_call(lexem)
    else
      raise "SYM-ERROR #{@lexer.get_line()}: Not a correct number #{lexem.val}|#{lexem.type}"
    end

    puts "Parser: Get_number returns #{value}" if @debug_mode
    return value
  end

  # Gets a certain index from a list or dict
  def get_by_index(lexem)
    r = /(?<name>[a-z][a-zA-Z0-9\_]*)\[(?<index>.+)\]/
    md = r.match(lexem.val)

    name = md["name"]
    index = md["index"]
    obj_ref = seek_storage(name)
    # The index might be an identifier, look it up if possible
    index_var = seek_storage(index, Variable, false)

    if obj_ref.class == SymList
      # Must be an integer, try to convert it if it's a string
      begin
        # If index is a variable, use it
        if index_var
          index = Integer(index_var)
        else
          index = Integer(index)
        end
      rescue Exception => e
        raise "SYM-ERROR #{@lexer.get_line} : Tried to access index "+
                                            "#{index}, but must be an integer"
      end
      # Index was converted to integer, return the value
      raise "SYM-ERROR #{@lexer.get_line} : Arrayindex out of bounds, "+
              "tried to access index #{index}, but array is defined to index "+
                    "#{obj_ref.list.length-1}" if index > obj_ref.list.length-1
                  
      return obj_ref.list[index]
    elsif obj_ref.class == SymDict
      begin
        if index_var
          e_index = index_var
        else
          # Evaluate index, it is currently only stored as string, could be anything
          temp_lexer = SymLexer.new(index)
          e_index = temp_lexer.get_next_lexem().val
        end
      rescue Exception => e
        raise "SYM-ERROR #{@lexer.get_line} : Could not evaluate "+
                                    "index #{index} in dictionary #{name}"
      end
      
      return obj_ref.get_value(e_index)
    else
      raise "SYM-ERROR #{@lexer.get_line} : Unable to find #{name}"
    end
  end

  def get_dict(lexem)
    puts "Parser: Entering get_dict()" if @debug_mode
    begins = 1
    ends = 0
    dict = {}
    # 0 = expteced key
    # 1 = expected seperator
    # 2 = expected value
    # 3 = expected comma
    expected = 0
    key = ""
    value = ""

    while true
      lex = @lexer.get_next_lexem()
      if lex.type == DictStart
        begins += 1
      elsif lex.type == DictEnd
        ends += 1
      end

      if begins == ends
        break
      end

      if lex.type == DictStart
        raise "SYM-ERROR #{@lexer.get_line()} : Dictionary expected "+
                                                  "seperator" if expected == 1
        raise "SYM-ERROR #{@lexer.get_line()} : Dictionary expected "+
                                                  "comma" if expected == 3
        obj = get_dict(lex)
        if expected == 0
          key = obj
          expected = 1

          ends += 1 # If successfully got dict, there was an end too
          @lexer.return_next()
          puts "Parser: Digested dictionary key #{key}" if @debug_mode
          next
        elsif expected == 2
          value = obj
          expected = 3
          dict[key] = value

          ends += 1 # If successfully got dict, there was an end too
          @lexer.return_next()
          puts "Parser: Digested dictionary value #{value}" if @debug_mode
          next
        end
      elsif lex.type == Terminator
        raise "SYM-ERROR #{@lexer.get_line()} : Unmatched dictionary tags"
      elsif lex.type == Comma
        if expected == 3
          expected = 0
        else
          raise "SYM-ERROR #{@lexer.get_line} : Unexpected comma in dictionary"
        end
      elsif lex.type == DictRelation
        if expected == 1
          expected = 2
        else
          raise "SYM-ERROR #{@lexer.get_line()}: Unexpected dict-seperator"
        end
      elsif expected == 0
        @lexer.return_previous()
        key = calculate_expr()
        expected = 1
        puts "Parser: Digested dictionary key #{key}" if @debug_mode
      elsif expected == 2
        @lexer.return_previous()
        value = calculate_expr()
        expected = 3
        dict[key] = value
        puts "Parser: Digested dictionary value #{value}" if @debug_mode
      else
        raise "SYM-ERROR #{@lexer.get_line} : Syntax error, invalid dictionary"
      end
    end

    puts "Parser: Dict created as #{dict}" if @debug_mode
    @lexer.return_previous()
    obj = SymDict.new(dict)
    puts "Parser: Returning from get_dict=#{dict}" if @debug_mode
    return obj
  end
  
  def get_list(lexem)
    puts "Parser: Entering get_list()" if @debug_mode
    begins = 1
    ends = 0
    list = []
    # This is used to detect errors with comma's
    # true=expected an element
    # false=expected comma
    element_allowed = true

    while true
      lex = @lexer.get_next_lexem()
      if lex.type == ListStart
        begins += 1
      elsif lex.type == ListEnd
        ends += 1
      end

      if begins == ends
        break
      end

      if lex.type == ListStart
        obj = get_list(lex)
        list << obj
      elsif lex.type == Terminator
        raise "SYM-ERROR #{@lexer.get_line()} : Unmatched list tags"

      elsif lex.type == Comma
        if element_allowed
          raise "SYM-ERROR #{@lexer.get_line()} : Duplicate comma"
        else
          element_allowed = true
        end
      elsif element_allowed
        element_allowed = false
        @lexer.return_previous()
        element = calculate_expr()
        if element.class != Lexem
          puts "Parser: Adding the #{element.class}=#{element} to list" if
                                                                       @debug_mode
          list << element
        end
        element_allowed = false
      else
        raise "SYM-ERROR #{@lexer.get_line()} : Invalid list, " +
                                        "expected comma to seperate elements"
      end
    end

    puts "Parser: List created as #{list}" if @debug_mode
    @lexer.return_previous() # Parser must find the terminator in main loop
    obj = SymList.new(list)
    return obj
  end

  def return_operator()
    puts "Parser: Entering return_operator()" if @debug_mode

    begin
      val = calculate_expr()
    rescue Exception=>e
      raise "SYM-ERROR #{@lexer.get_line()}: Could not " +
                                  "successfully return from #{@in_fun.name}"
    end

    raise "SYM-ERROR #{@lexer.get_line()}: Tried to return"  +
                                " outside function or method" if not @in_fun

    # Perform typecheck
    # Void=Don't perform typecheck
    # Int is alias for fixnum, so allow them to be interchanged
      raise "SYM-ERROR #{@lexer.get_line()}: Invalid returntype, can't" +
        "  return a #{val.class}, expected to return #{@in_fun.returntype}" if
                                            @in_fun.returntype != Void and
                                            @in_fun.returntype != val.class and
                        not (@in_fun.returntype == Int and val.class == Fixnum)

    @return_value = val
    @break = true # Break main iteration when possible
    puts "Parser: return value set to #{@return_value }" if @debug_mode
  end
  
  def class_init(lexem)
    puts "Parser: Creating an object with #{lexem.val}" if @debug_mode

    # Use regexp to find each seperate part
    match = /(.*)\.(.+)\((.*)\)/.match(lexem.val)
    name = match[1]
    args = match[3]
    
    class_obj = seek_storage(name, Class)
    raise "SYM-ERROR #{@lexer.get_line()}: Couldn't find " +
        "init constructor in class #{name}" if not class_obj.methods["init"]

    # Create the object
    object = SymObject.new(name, class_obj.variables, class_obj.methods)
    object.scope.store_object("self", object)

    # Add object scope
    object_scope = object.scope
    @scopes << object_scope

    # Re-use the call function code, but it expects a lexem object
    # So we need to trick it
    l = Lexem.new(FunctionCall, "init"+"("+args+")")
    call_function(l)

    @scopes.delete(object_scope)
    return object
  end

  def call_method(lexem, objname)
    puts "Parser: Entering call_method()" if @debug_mode

    # Look up the variables used for recursion
    max_rec = seek_storage("sym_max_recursion", Variable, false)
    c_rec = seek_storage("sym_current_recursion", Variable, false)
    return @return_value if c_rec >= max_rec and not max_rec == -1

    # Use regexp to get each part
    match = /\.(.+?)\((.*)\)/.match(lexem.val)
    name = match[1]
    args = match[2]

    obj_ref = seek_storage(objname, SymObject)

    method_ref = obj_ref.methods[name]
    raise "SYM-ERROR #{@lexer.get_line()}: Method #{name} could not " +
                          "be found in object #{objname}" if not method_ref

    # Object scope
    objscope = obj_ref.scope
    @scopes << objscope
    @scopes[-1].store_object("self", obj_ref)

    # Set current function to self, restore old reference afterwards
    old_ref = @in_fun
    @in_fun = method_ref

    # Run the method
    puts "Parser: Calling method #{name} in object #{objname}" if @debug_mode
    # Re-use the code in call_function, it expects a lexem object
    l = Lexem.new(FunctionCall, name+"(#{args})")
    call_function(l)

    @scopes.delete(objscope)
    return @return_value
  end

  # NOTE: Cluttered code, this is due to the fact that
  # it's re-used in method calls and class init
  def call_function(lexem)
    puts "Parser: Functioncall with #{lexem.val}" if @debug_mode

    # Make sure we havn't went above the maximum recursion depth
    # -1 means don't have a limit
    max_rec = seek_storage("sym_max_recursion", Variable, false)
    c_rec = seek_storage("sym_current_recursion", Variable, false)
    return @return_value if c_rec >= max_rec and not max_rec == -1

    re = /(.+)\((.*)\)/
    name = lexem.val.split(re)[1]
    args = lexem.val.split(re)[2]   # A string of all arguments
    argument_list = nil
    argument_list = args.split(",") if args # An array of all arguments

    fun_ref = seek_storage(name, Function) # Function reference
    raise "SYM-ERROR #{@lexer.get_line()}: Function #{name} is not defined" if
                                                                  not fun_ref
    arguments = {}
    argval = nil
    index = 0
    
    for argname in fun_ref.arguments.keys
      
      begin
        # First look for objects of this name, if that doesn't work
        # Evaluate it as a literal
        argval = seek_storage(argument_list[index], Variable, false)
        argval = eval(argument_list[index].to_s) if not argval
      rescue Exception => e
        raise "SYM-ERROR #{@lexer.get_line()}: Invalid functioncall." +
                    "Could not evaluate argument #{argument_list[index].to_s}"
      end
      raise "SYM-ERROR #{@lexer.get_line()}: Invalid functioncall." +
                                " Argument #{argname} not supplied or " +
                                      "it was something invalid" if not argval

      # Raise error if it's not the right kind of variable
      # Don't perform typecheck if it's of void type
      # Int is just an alias for fixnum, so allow either one
      if fun_ref.arguments[argname] != Void and 
          fun_ref.arguments[argname] != argval.class and
            not (fun_ref.arguments[argname] == Int and argval.class == Fixnum)
          
        # If Int is expected the variable might be a Fixnum
        # Because that's how integers are stored internally in ruby
        raise "SYM-ERROR #{@lexer.get_line()}: Invalid functioncall."+
          " Argument #{argname} expected type #{fun_ref.arguments[argname] if
            fun_ref.arguments[argname] != Fixnum}" +
            "#{Int if fun_ref.arguments[argname] == Fixnum} but "+
            "got #{argval.class if argval.class != Fixnum}"+
            "#{Int if argval.class == Fixnum}"
      end
      arguments[argname] = argval
      index += 1
    end

    # Create a new local scope for the function, and put the arguments in it
    newscope = Scope.new(arguments, @debug_mode)
    @scopes << newscope

    puts "Parser: Calling function #{name}" if @debug_mode

    # Set current function to self, restore old reference afterwards
    old_ref = @in_fun
    @in_fun = fun_ref
    @scopes[0].storage["sym_current_recursion"] += 1

    self.parse(fun_ref.code) # Call the function code
    # Reset
    @in_fun = old_ref
    @scopes.delete(newscope)
    @scopes[0].storage["sym_current_recursion"] -= 1
    return @return_value
  end

  def class_def(lexem)
    puts "Parser: Entering class_def()" if @debug_mode
    
    # Terminators(newlines) are allowed in infinite amounts in a class def
    # This function is used to removes them
    def get_lex()
      lex = @lexer.get_next_lexem()
      if lex.type != Terminator
        return lex
      else
        return get_lex()
      end
    end

    # Class name
    name_lex = lexem
    raise "SYM-ERROR #{@lexer.get-line}: " if name_lex.type == Variable
    name = name_lex.val.slice(2..-1)

    # Class variables
    variables = {}
    next_lex = get_lex()
    not_allowed = ["return", "SYM_VERSION", "sym_max_recursion",
                       "sym_current_recursion", "true", "false"]
    while (next_lex.type != FunctionDef and next_lex.type != ClassEnd)
      raise "SYM-ERROR #{@lexer.get_line}: Can not create reserved "+
             "identifier #{next_lex.val}" if not_allowed.include?(next_lex.val)
      raise "SYM-ERROR #{@lexer.get_line()}: Can not "+
        "create reserved identifier return " if next_lex.type == ReturnOperator
      
      variables[next_lex.val] = 0
      next_lex = get_lex()
    end




    puts "Parser: Attempting to read class methods list" if @debug_mode
    # Methods
    methods = {}
    while (next_lex.type == FunctionDef)
      ans = function_def()
      # name, ret, code, args, scope_ref
      obj = Function.new(ans[0], ans[1], ans[2], ans[3], ans[4])
      methods[obj.name] = obj
      next_lex = get_lex()
    end

    if @debug_mode 
      puts "Parser: Creating a class:"
      puts "\tName: #{name}"

      # Class variables
      if variables.length > 0
        puts "\tVariables: "
        for var in variables.keys
          puts "\t\t#{var}"
        end
      end

      # Methods, their returntype and possible arguments
      if methods.length > 0
        puts "\tMethods: "
        for fun_name in methods.keys
          puts "\t\t#{fun_name}:"
          puts "\t\t\tReturns: #{methods[fun_name].returntype} "
          if methods[fun_name].arguments.length > 0
            puts "\t\t\tArgs:"
            for args in methods[fun_name].arguments.keys
              puts "\t\t\t\t#{ methods[fun_name].arguments[args]}, #{args}"
            end
          end
        end
      end
      
    end

    # Create the class and save it
    c_obj = Class.new(name, variables, methods)
    @scopes[-1].store_object(name, c_obj)
    @lexer.add_datatype(name)
  end

  # Creates a function and stores it in scope.
  # Gets different lexems from lexer depending on arguments.
  # If there are no arguments, it will get "FunctionCall" lexem, etc...
  def function_def()
    puts "Parser: Entering function_def()" if @debug_mode
    # Terminators(newlines) are allowed in infinite amounts in a function def
    # This function is used to removes them
    def get_lex()
      lex = @lexer.get_next_lexem()
      if lex.type != Terminator
        return lex
      else
        return get_lex()
      end
    end

    # Returntype:
    ret_lex = get_lex()
    name_lex = ""
    # If there was no returntype specified, this lex is the name instead
    # We also default to type Void
    if ret_lex.type != DataType
      name_lex = ret_lex
      ret_lex = Lexem.new(DataType, DataType.new(Void))
    else
      name_lex = get_lex()
    end
    name = name_lex.val

    # Argument hashes
    # Key=name, as string
    # Value=type(Int, String etc..), Void=Don't force anything
    args = {}
    # Recognised as functioncall if there's no arguments
    if name_lex.type == FunctionCall
      reg = /\A(.+)\((.*)\)/
      md = reg.match(name_lex.val)

      # Raise error if there was no match
      if md
        name = md[1]

        # If there was some (none type specified) arguments, save them (as void)
        if md[2]
          for f_arg in md[2].split(",")
            args[f_arg] = Void
          end
        end
      
      else
        raise "SYM-ERROR #{@lexer.get_line()}: Invalid functiondef"
      end
    else
      # L paranthesis:
      lp_lex = get_lex()

      # Arguments:
      previous = lp_lex
      current_lex = nil

      while (true) # Raise's implement syntax check
        current_lex = get_lex()

        if current_lex.type == Variable
          if previous.val.class == DataType
            args[current_lex.val] = previous.val.datatype
          else
            args[current_lex.val] = Void
          end
          
          previous = current_lex
        elsif current_lex.type == DataType
          previous = current_lex
        elsif current_lex.type == Comma
          raise "SYM-ERROR #{@lexer.get_line()}: Invalid functiondef. Proper syntax is: \n$/Returntype name(Arguments(Datatype Name)*)\n  Body\n\\$\n * =Arguments are optional, leave as () for no arguments" if previous.type != Variable
        elsif current_lex.type == RParanthesis # We reached the end of the definition
          # Raise an error if the last lexem was a datatype, it needs to have a name
          raise "SYM-ERROR #{@lexer.get_line()}: Invalid functiondef. Proper syntax is: \n$/Returntype name(Arguments(Datatype Name)*)\n  Body\n\\$\n * =Arguments are optional, leave as () for no arguments" if previous.type != Variable and previous.type != LParanthesis
          break
        else
          raise "SYM-ERROR #{@lexer.get_line()}: Invalid functiondef. Proper syntax is: \n$/Returntype name(Arguments(Datatype Name)*)\n  Body\n\\$\n * =Arguments are optional, leave as () for no arguments"
        end
    end
    end

    # Perform check for reserved words in function arguments
    for a in args.keys
      not_allowed = ["return", "SYM_VERSION", "sym_max_recursion",
                                  "sym_current_recursion", "true", "false"]

      raise "SYM-ERROR : Can not use reserved identifier #{a} as argument" if
                                                not_allowed.include?(a)
    end

    
    # Code
    code = @lexer.get_function_body() 
    ret = ret_lex.val.datatype
    scope_ref = get_fun_scope()

    if @debug_mode
      puts "Parser: Creating function, using:"
      puts "  Returntype: #{ret}"
      puts "  Name: #{name}"
      
      puts "  Arguments: "
      for argname in args.keys
        print "    #{argname}="
        puts args[argname]
      end

      puts "  Code: '#{@lexer.code_left(code)}'"
    end

    return [name, ret, code, args, scope_ref]
  end

  def get_fun_scope()
    result = []
    for scope in @scopes
      result << scope
    end
    return result
  end

  def input_operator()
    local_debug = false
    puts "Parser: Entering input_operator()" if @debug_mode or local_debug
    lexem = @lexer.get_next_lexem()

    if lexem.type == Variable
      val = STDIN.gets().slice(0..-2)
	  @scopes[-1].store_object(lexem.val, val)
    elsif lexem.type == Constant
      val = STDIN.gets().slice(0..-2)
	  constant_exists = @scopes[-1].get_name(lexem.val)
	  if (constant_exists)
        raise "SYM-ERROR: Constant #{lexem.val} already in use"
	  else
	    @scopes[-1].store_object(lexem.val, val, true)
	  end
    else
      raise "SYM-ERROR #{@lexer.get_line()}: Invalid lexem" +
                        " , expected variable/constant '#{lexem.val}'"
    end

    puts "Parser: Assigning the #{lexem.type}" +
                      " #{lexem.val} to #{val}" if @debug_mode or local_debug
  end

  def modifier_operator(lexem_val)
    puts "Parser: Entering modifier_operator()" if @debug_mode
    puts "Parser: lexem_val: " + lexem_val.to_s if @debug_mode
    varname = @lexer.get_grand()

    if varname.type == Variable
      temp_var = seek_storage(varname.val)
      val = calculate_expr()

	  check_numeric_error(temp_var, val, lexem_val) if lexem_val != "*="
	  
      if (lexem_val == "+=")
        plus_val = temp_var.to_i + val
      elsif (lexem_val == "-=")
        plus_val = temp_var - val
      elsif (lexem_val == "*=")
		raise "SYM-ERROR #{@lexer.get_line()}: Invalid operation. " +
			" Right hand operand must be numeric" if !is_numeric?(val)
        plus_val = temp_var * val
      elsif (lexem_val == "/=")
		raise "SYM-ERROR #{@lexer.get_line()}: Can't divide variable with zero" if val.ceil == 0
        plus_val = temp_var / val
      elsif (lexem_val == "%=")
        plus_val = temp_var % val
      elsif (lexem_val == "**=" || lexem_val == "^=")
        plus_val = temp_var ** val
      else
        raise "SYM-PARSE-ERROR OCCURRED: Contact Sym Team. Report error code #1"
      end

	  @scopes[-1].store_object(varname.val, plus_val)
    elsif varname.type == Constant
      raise "Error: Constant #{varname.val.to_s} can not be respecified"
    else
      raise "Error: Invalid lexem for assignment #{varname.type}"
    end
  end

  def assignment_operator()
    puts "Parser: Entering assignment_operator()" if @debug_mode
    varname = @lexer.get_grand()
    val = calculate_expr()

    # Check if it already exists in some scope
    # If so, assign the value to this variable instead of creating a new one
    answ = seek_storage(varname.val, varname.type, false)
    if answ
      @scopes[@return_level].store_object(varname.val, val, varname.type == Constant)
    else # Doesn't exist already, just put it in the local scope
      # Also make sure the name is valid variable/constant
      if varname.type == Variable or varname.type == Constant
        @scopes[-1].store_object(varname.val, val)
      else
        raise "SYM-ERROR #{@lexer.get_line()}: Invalid lexem for"+
                                            " assignment #{@varname.type}"
      end
    end
  end

  def print_operator(lexem = nil)
    puts "Parser: Entering print_operator()" if @debug_mode
    value = calculate_expr()

    # Objects are treated specially
    if value.class == SymObject
      if value.methods.include?("__print__")
        # The object has the special __print__ method, call it
        objname = find_name(value)
        str = "#{objname}.__print__()"
        self.parse(str)
      else
        raise "SYM-ERROR #{@lexer.get_line()}: Tried to print"+
                                      " object. Define the __print__() method."
      end
    elsif value.class == SymList or value.class == SymDict
      value.print_self()

    else # Everything not an object
      # Replace escapes in strings
      if (value.class == String)
        value = value.gsub(/\\n/, "\n")
        value = value.gsub(/\\t/, "\t")

        value = value.gsub(/\\r/, "\r")
        value = value.gsub(/\\s/, "\s")
        value = value.gsub(/\\\\/, "\\") # Double backspace
      end

      # Don't print anything if silent mode is enabled
      # Highlight output in debug mode
      if lexem and lexem.val
        print "--SYM-->" if @debug_mode and not @silent_mode
        print value if not @silent_mode
        print "<-SYM---" if @debug_mode and not @silent_mode
      else
        puts "--SYM-->" if @debug_mode and not @silent_mode
        puts value if not @silent_mode
        puts "<-SYM---" if @debug_mode and not @silent_mode
      end

      end
  end
  
  def is_numeric?(arg)
    return arg.is_a? Numeric
  end
  
  def check_numeric_error(factor1, factor2, lexem_val)
    def has_all_numeric(*args)
	  for arg in args
		return false if not is_numeric?(arg)
	  end
	  return true
	end
	
	raise "SYM-ERROR: Operation #{lexem_val} invalid on '#{factor1.to_s}' and '#{factor2.to_s}'" if not has_all_numeric(factor1, factor2)
  end
  def conditional_statement()
    local_debug = false
    puts "##############################" if local_debug or @debug_mode
    puts "Entering conditional_statement" if local_debug or @debug_mode

    if_code = ""
	if_condition = ""
    else_if_code = []
    else_code = ""
    accumulated_code = ""

    save_to_if_code = "saving code to if code"
    save_to_else_if_code_first = "saving code to else if code (first)"
    save_to_else_if_code_second = "saving code to else if code (second)"
    save_to_else_code = "saving code to else code"

    save_to_code = save_to_if_code
    saved_to_last = save_to_code

    # Used to check multiple appearances of "else" in a flat if structure,
    # which is strictly unallowed
    else_code_saved = false
    identity = nil

    if_level = 1

    while (if_level != 0)
      puts "if_level: #{if_level.to_s}" if local_debug or @debug_mode
      code, condition, next_part =
			@lexer.get_code_body(/(\\\?\/)|(\\\?)|(\?\/)|(\\\*)/)
      puts "ret code\n\"#{code.to_s}\"" if local_debug or @debug_mode
      puts "ret condition\n\"#{condition.to_s}\"" if local_debug or @debug_mode
      puts "ret next_part\n\"#{next_part.to_s}\"" if local_debug or @debug_mode

      # Keep track where code was saved last time.
      # If it has changed set code to accumulated_code
      if (saved_to_last != save_to_code)
        puts "changing saving. Now #{save_to_code}" if local_debug or @debug_mode
        puts "saved_to_last: #{saved_to_last}" if local_debug or @debug_mode
        if (saved_to_last == save_to_if_code)
          if_code = [accumulated_code, identity]
          puts "saved if code: #{if_code.to_s}" if local_debug or @debug_mode
        elsif (saved_to_last == save_to_else_code)
          else_code = [accumulated_code, identity]
          puts "saved else_code: #{else_code.to_s}" if local_debug or @debug_mode
        elsif (saved_to_last == save_to_else_if_code_first or
              saved_to_last == save_to_else_if_code_second)
          else_if_code << [accumulated_code, identity]
        end

        # Empty accumulated_code so it can be reused
        accumulated_code = ""
	  else
		if_condition = condition
      end

      if (if_level == 1)
		newscope = Scope.new({"identity" => "0", }, @debug_mode)
		@scopes << newscope
        self.parse("identity = " + condition) if condition != ""
		identity = @scopes[-1].get_name("identity")
		@scopes.delete(newscope)
        puts "condition was: #{identity.to_s}" if local_debug or @debug_mode
      end

      saved_to_last = save_to_code

      # If level of code is greater than 1 add IfParts that would otherwise
      # disappear when getting code from get_code_body()
      if (if_level > 1)
        if (next_part == "?/" or next_part == "\\?") # IfPart or EndIfPart
          accumulated_code += "?/ " if condition != ""
          accumulated_code += condition
        elsif (next_part == "\\*") # ElsePart
          accumulated_code += "?/ " if condition != ""
          accumulated_code += condition
        elsif (next_part == "\\?/") # ElseIfPart
          accumulated_code += "?/ " if condition != ""
          accumulated_code += condition
        end
      end

	  accumulated_code += code if code != nil

	  # Increase or decrease "if level" if necessary. Also add IfParts to
	  # the accumulated code, to make the code correct
	if (next_part == "?/") # IfPart
      puts "Next part was IfPart" if local_debug or @debug_mode
      if_level += 1
    elsif (next_part == "\\?") # EndIfPart
      puts "Next part was EndIfPart" if local_debug or @debug_mode
      accumulated_code += "\\?" if if_level != 1
      if_level -= 1
    elsif (next_part == "\\*") # ElsePart
      puts "Next part was ElsePart" if local_debug or @debug_mode
      accumulated_code += "\\*" if if_level != 1
    elsif (next_part == "\\?/") # ElseIfPart
      puts "Next part was ElseIfPart" if local_debug or @debug_mode
      accumulated_code += "\\" if if_level != 1
    else
      raise "SYM-ERROR #{@lexer.get_line()} #{condition.to_s}: Undefined if structure"
	end

	# If the if structure is flat (i e the if level is == 1)
    # change save_to_code to else or else if.
	  if (if_level == 1)
		if (next_part == "\\*")
		  raise "SYM-ERROR ?\/ #{if_condition.to_s}: Multiple appearances of else part" if else_code_saved
          puts "saving code to else..." if local_debug or @debug_mode
          save_to_code = save_to_else_code
          else_code_saved = true
        elsif (next_part == "\\?/")
          # Shift between variables to make sure that code is saved correctly
          save_to_code = save_to_else_if_code_first ==
          save_to_code ? save_to_else_if_code_second : save_to_else_if_code_first
        end
	  end

	  puts "acc code: #{accumulated_code.to_s}" if local_debug or @debug_mode
  end

	if (saved_to_last == save_to_if_code)
	  if_code = [accumulated_code, identity]
	  puts "saved if code: #{if_code.to_s}" if local_debug or @debug_mode
	elsif (saved_to_last == save_to_else_code)
	  else_code = [accumulated_code, identity]
	  puts "saved else_code: #{else_code.to_s}" if local_debug or @debug_mode
	elsif (saved_to_last == save_to_else_if_code_first or
      saved_to_last == save_to_else_if_code_second)
	  else_if_code << [accumulated_code, identity]
	end

	# Print all different codes
	puts "##############################" if local_debug or @debug_mode
	puts "if_code:\"\n#{if_code[0].to_s}\"" if local_debug or @debug_mode
	puts "if_code condition: #{if_code[1].to_s}\"" if local_debug or @debug_mode

	else_if_code.each_index do |i|
	  puts "elseif_code:\n\"#{else_if_code[i][0]}\"" if local_debug or
      @debug_mode
	  puts "elseif_condition: #{else_if_code[i][1]}" if local_debug or
      @debug_mode
	end
	puts "else_code:\"\n#{else_code[0].to_s}\"" if local_debug or @debug_mode
	puts "------------------------------" if local_debug or @debug_mode

	code_to_run = nil
	# Choose the code to run according to each if-parts identity values
	if (if_code[1] == true)
	  code_to_run = if_code[0]
	else
	  # step through each else if code and see if its condition is true
	  else_if_code.each_index do |i|
		if (else_if_code[i][1] == true)
		  code_to_run = else_if_code[i][0]
		  break
		end
	  end
	  # if there were no true conditions in "if" or "else if", set code_to_run
	  # to "else" code, if there are any
	  if (code_to_run == nil)
		code_to_run = else_code[0] if else_code[0] != nil
	  end
	end
	code_to_run = code_to_run.strip if code_to_run
	puts "final code:\n\"#{code_to_run.to_s}\"" if local_debug or @debug_mode
	self.parse(code_to_run) if code_to_run != "" and code_to_run != nil
  end
  
 def iterative_for_statement()
	local_debug = false

	puts "Entering iterative_for_statement" if local_debug or @debug_mode

	accumulated_code = ""
	condition_saved = false
	saved_condition = ""

	iterative_for_level = 1
	while (iterative_for_level != 0)
		code, condition, next_part = @lexer.get_code_body(/(\.\.\/)|(\\\.\.)/)
		puts "returned code\n\"#{code.to_s}\"" if local_debug or @debug_mode
		puts "ret condition\n\t\"#{condition.to_s}\"" if local_debug or @debug_mode
		puts "ret next_part\n\t\"#{next_part.to_s}\"" if local_debug or @debug_mode

		# static check - only check once
		if not condition_saved
		    valid_num = /[1-9][0-9]*|0/

			minimal_parts = /\A(#{valid_num})\s+(#{valid_num})\s*$/.match(condition)
			maximal_parts = nil
			if minimal_parts
				raise "SYM-ERROR ../ #{condition.to_s}: Invalid parts of for-loop" if not minimal_parts
				min_val = minimal_parts.captures[0].to_i
				max_val = minimal_parts.captures[1].to_i
				local_val = nil
				step_val = 0
			else
				maximal_parts = /\A(#{valid_num})\s+(#{valid_num})\s+([a-z][a-zA-Z0-9_]*)(\s+(([1-9][0-9]*$)|0))?/.match(condition)
				if maximal_parts
					min_val = maximal_parts.captures[0].to_i
					max_val = maximal_parts.captures[1].to_i
					local_val = maximal_parts.captures[2].to_s
					step_val = maximal_parts.captures[3]
				end
				raise "SYM-ERROR ../ #{condition.to_s}: Invalid parts of for-loop" if not maximal_parts
			end

			if not step_val
			  match_invalid_step_val = /\A\d+\s+\d+\s+[a-z][a-zA-Z0-9_]*\s(\s*.*)?/.match(condition)
			  if match_invalid_step_val
				invalid_step_val = match_invalid_step_val.captures[0].to_s
				raise "SYM-ERROR #{@lexer.get_line().strip()} #{condition.to_s}: " +
				"Invalid step value --> #{invalid_step_val.to_s} <--"
			  end
			end
			step_val = step_val.to_i

			puts "min_val: #{min_val.to_s}" if local_debug or @debug_mode
			puts "max_val: #{max_val.to_s}" if local_debug or @debug_mode
			puts "local_val: #{local_val.to_s}" if local_debug or @debug_mode
			puts "step_val: #{step_val.to_s}" if local_debug or @debug_mode
			raise "SYM-ERROR #{@lexer.get_line()} #{condition.to_s}: Illegal values to" +
				" for loop. Min val has to be lower than or equal to Max val" if
				min_val.to_i > max_val.to_i
			condition_saved = true
		end

		if (iterative_for_level > 1)
		  if (next_part == "../" or next_part == "\\..")
			accumulated_code += "../" if condition != ""
			accumulated_code += condition
		  end
		end

		accumulated_code += code if code != nil

		if (next_part == "../") # ForPart
		  puts "Next part was ForPart" if local_debug or @debug_mode
		  iterative_for_level += 1
		elsif (next_part == "\\..") # EndForPart
		  puts "Next part was EndForPart!" if local_debug or @debug_mode
		  accumulated_code += "\\.." if iterative_for_level != 1
		  iterative_for_level -= 1
		else
		  raise "SYM-ERROR ../ #{condition}: Unfinished for structure. \\.. missing"
		end

	end

	puts "Code at end:\"#{accumulated_code.to_s}\"" if local_debug or @debug_mode
	step_counter = 0
	@loop_level += 1
	min_val.upto(max_val) do |i|
		@scopes[-1].store_object(local_val, i) if local_val != nil
		break if step_val >= max_val
		if step_counter == 0
		  self.parse(accumulated_code)
		  break if @break_loop
		  step_counter = step_val+1
		end
		step_counter -= 1 if step_counter > 0
	end
	@loop_level -= 1
	@break_loop = false
	@scopes[-1].unstore_object(local_val) if local_val != nil
	puts "end of iterative_for_statement" if local_debug or @debug_mode
  end
  
   def iterative_while_statement()
	local_debug = false
	iterative_while_level = 1
	puts "			Entering iterative_while_statement" if local_debug or @debug_mode
	puts "			while_level is : #{iterative_while_level}" if local_debug

	accumulated_code = ""
	condition_saved = false
	saved_condition = ""
	while (iterative_while_level != 0)
	  code, condition, next_part = @lexer.get_code_body(/(\|\/)|(\\\|)/)
	  puts "ret code\n\"#{code.to_s}\"" if local_debug or @debug_mode
	  puts "ret condition\n	\"#{condition.to_s}\"" if local_debug or @debug_mode
	  puts "ret next_part\n	\"#{next_part.to_s}\"" if local_debug or @debug_mode

	  if not condition_saved
		saved_condition = condition
		condition_saved = true
		raise "SYM-ERROR #{@lexer.get_line()}: Invalid while condition" if saved_condition == ""
	  end

	  if (iterative_while_level > 1)
		if (next_part == "|/" or next_part == "\\|")
		  accumulated_code += "|/" if condition != ""
		  accumulated_code += condition
		end
	  end

	  accumulated_code += code if code != nil

	  if (next_part == "|/") # While
		puts "Next part was WhilePart" if local_debug or @debug_mode
		iterative_while_level += 1
	  elsif (next_part == "\\|") # EndWhile
		puts "Next part was EndWhilePart" if local_debug or @debug_mode
		accumulated_code += "\\|" if iterative_while_level != 1
		iterative_while_level -= 1
	  else
		raise "SYM-ERROR #{@lexer.get_line()} #{condition}:" +
		" Unfinished while structure. \\| missing"
	  end

	end

	puts "Acc code at end:\"#{accumulated_code.to_s}\"" if local_debug or
	  @debug_mode
	puts "Saved condition: \"#{saved_condition.to_s}\"" if local_debug or
	  @debug_mode
	accumulated_code = accumulated_code.strip

	@loop_level += 1
	while true
	  newscope = Scope.new({"identity" => "0", }, @debug_mode)
	  @scopes << newscope
	  self.parse("identity = " + saved_condition)
	  identity = @scopes[-1].get_name("identity")
	  @scopes.delete(newscope)
	  if identity
		self.parse(accumulated_code) if accumulated_code != "" and not @break_loop
	  end
	  break if not identity or @break_loop
	end
	@break_loop = false
	puts "Ending iterative_while_statement" if local_debug or @debug_mode
	@loop_level -= 1
end
end